import warnings

from sqs_workers.codecs import DEFAULT_CONTENT_TYPE
from sqs_workers.utils import adv_bind_arguments


class AsyncTask(object):
    def __init__(self, queue, job_name, processor):
        self.queue = queue
        self.job_name = job_name
        self.processor = processor
        self.__doc__ = processor.__doc__

    def __call__(self, *args, **kwargs):
        warnings.warn(
            "Async task {0.queue_name}.{0.job_name} called synchronously".format(self)
        )
        return self.processor(*args, **kwargs)

    def __repr__(self):
        return "<%s %s.%s>" % (self.__class__.__name__, self.queue.name, self.job_name)

    def delay(self, *args, **kwargs):
        _content_type = kwargs.pop("_content_type", DEFAULT_CONTENT_TYPE)
        _delay_seconds = kwargs.pop("_delay_seconds", None)
        _deduplication_id = kwargs.pop("_deduplication_id", None)
        _group_id = kwargs.pop("_group_id", None)
        kwargs = adv_bind_arguments(self.processor, args, kwargs)
        return self.queue.add_job(
            self.job_name,
            _content_type=_content_type,
            _delay_seconds=_delay_seconds,
            _deduplication_id=_deduplication_id,
            _group_id=_group_id,
            **kwargs
        )

    def bake(self, *args, **kwargs):
        """
        Create a baked version of the async task, which contains the reference
        to a queue and task, as well as all arguments which needs to be passed
        to it.
        """
        return BakedAsyncTask(self, args, kwargs)


class BakedAsyncTask(object):
    def __init__(self, async_task, args, kwargs):
        self.async_task = async_task
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        self.async_task(*self.args, **self.kwargs)

    def delay(self):
        self.async_task.delay(*self.args, **self.kwargs)

    def __repr__(self):
        return "BakedAsyncTask(%r, ...)" % self.async_task



from sqs_workers.backoff_policies import (  # noqa: F401
    DEFAULT_BACKOFF,
    IMMEDIATE_RETURN,
    ConstantBackoff,
    ExponentialBackoff,
)
from sqs_workers.memory_sqs import MemorySession  # noqa: F401
from sqs_workers.queue import JobQueue, RawQueue  # noqa: F401
from sqs_workers.sqs_env import SQSEnv  # noqa: F401
from sqs_workers.sqs_manage import (  # noqa: F401
    create_fifo_queue,
    create_standard_queue,
    delete_queue,
)



import logging
from functools import partial
from typing import TYPE_CHECKING

import attr

from sqs_workers import RawQueue

if TYPE_CHECKING:
    from sqs_workers.queue import GenericQueue


logger = logging.getLogger(__name__)


@attr.s
class DeadLetterQueue(RawQueue):
    """
    Queue to push back messages to the upstream.

    While processing the queue, the processor takes every message and push it
    back to the upstream queue with a hard-coded delay of 1 second.

    Usage example:

    >>> from sqs_workers import JobQueue
    >>> from sqs_workers.shutdown_policies IdleShutdown
    >>> from sqs_workers.deadletter_queue import DeadLetterQueue
    >>> env = SQSEnv()
    >>> foo = env.queue("foo")
    >>> foo_dead = env.queue("foo_dead", DeadLetterQueue.maker(foo))
    >>> foo_dead.process_queue(shutdown_policy=IdleShutdown(10))

    This code takes all the messages in foo_dead queue and push them back to
    the foo queue. Then it waits 10 seconds to ensure no new messages appear,
    and quit.
    """

    upstream_queue = attr.ib(default=None)  # type: GenericQueue

    @classmethod
    def maker(cls, upstream_queue, **kwargs):
        processor = PushBackSender(upstream_queue)
        return partial(
            cls, processor=processor, upstream_queue=upstream_queue, **kwargs
        )


@attr.s
class PushBackSender(object):

    upstream_queue = attr.ib(default=None)  # type: GenericQueue

    def __call__(self, message):
        kwargs = {
            "MessageBody": message.body,
            "MessageAttributes": message.message_attributes or {},
        }

        if self.upstream_queue.name.endswith(".fifo"):
            kwargs.update(
                {
                    "MessageDeduplicationId": message.attributes[
                        "MessageDeduplicationId"
                    ],
                    "MessageGroupId": message.attributes["MessageGroupId"],
                }
            )
        else:
            kwargs.update({"DelaySeconds": 1})
        self.upstream_queue.get_queue().send_message(**kwargs)



import importlib
from typing import Any

from werkzeug.utils import bind_arguments, validate_arguments


def adv_validate_arguments(callback, args, kwargs):
    """
    "Advanced version" of Werkzeug's "validate_arguments" which doesn't modify
    passed args and kwargs

    :return: (args, kwargs) to pass to original function
    """
    bind_args = list(args)
    bind_kwargs = kwargs.copy()
    arguments, keyword_arguments = validate_arguments(callback, bind_args, bind_kwargs)
    return arguments, keyword_arguments


def adv_bind_arguments(callback, args, kwargs):
    """
    "Advanced version" of Werkzeug's "bind_arguments" which doesn't modify
    passed args and kwargs
    """
    bind_args = list(args)
    bind_kwargs = kwargs.copy()
    keyword_arguments = bind_arguments(callback, bind_args, bind_kwargs)
    return keyword_arguments


def string_to_object(string):
    # type: (str) -> Any
    """
    Convert full path string representation of the object to object itself.
    """
    chunks = string.rsplit(".", 1)
    if len(chunks) != 2:
        raise RuntimeError(
            "{} doesn't represent a full module path to object".format(string)
        )
    module_name, object_name = chunks
    mod = importlib.import_module(module_name)
    return getattr(mod, object_name)


def instantiate_from_string(string, **init_kwargs):
    """
    Create an object from a classname string and init kwargs.
    """
    class_ = string_to_object(string)
    return class_(**init_kwargs)


def instantiate_from_dict(options, maker_key="maker", **extra_init_kwargs):
    """
    Create an object from options with classname and init kwargs.
    """
    kwargs = options.copy()
    classname_value = kwargs.pop(maker_key)
    init_kwargs = dict(kwargs, **extra_init_kwargs)
    return string_to_object(classname_value)(**init_kwargs)



import datetime

now = datetime.datetime.utcnow


class NeverShutdown(object):
    """
    Never shutdown the worker
    """

    def update_state(self, batch_processing_result):
        pass

    def need_shutdown(self):
        return False

    def __repr__(self):
        return "NeverShutdown()"


class IdleShutdown(object):
    """
    Shutdown worker if it's idle for certain time (set in seconds)
    """

    def __init__(self, idle_seconds):
        self.idle_seconds = idle_seconds
        self._idle_delta = datetime.timedelta(seconds=idle_seconds)
        self._is_idle = False
        self._last_seen = now()

    def update_state(self, batch_processing_result):
        """
        Update internal state of the shutdown policy
        """
        if batch_processing_result.total_count() == 0:
            self._is_idle = True
        else:
            self._is_idle = False
            self._last_seen = now()

    def need_shutdown(self):
        if not self._is_idle:
            return False
        return now() - self._last_seen >= self._idle_delta

    def __repr__(self):
        return "IdleShutdown({})".format(self.idle_seconds)


class MaxTasksShutdown(object):
    """
    Shutdown worker if it executed more than max_tasks in total (both
    successfully or with error)
    """

    def __init__(self, max_tasks):
        self.max_tasks = max_tasks
        self._tasks = 0

    def update_state(self, batch_processing_result):
        self._tasks += batch_processing_result.total_count()

    def need_shutdown(self):
        return self._tasks >= self.max_tasks

    def __repr__(self):
        return "MaxTasksShutdown({})".format(self.max_tasks)


class OrShutdown(object):
    """
    Return True if any of conditions of policies is met
    """

    def __init__(self, *policies):
        self.policies = policies

    def update_state(self, batch_processing_result):
        for p in self.policies:
            p.update_state(batch_processing_result)

    def need_shutdown(self):
        for p in self.policies:
            if p.need_shutdown():
                return True
        return False

    def __repr__(self):
        repr_policies = ", ".join([repr(p) for p in self.policies])
        return "OrShutdown({})".format(repr_policies)


class AndShutdown(object):
    """
    Return True if all of conditions of policies are met
    """

    def __init__(self, *policies):
        self.policies = policies

    def update_state(self, batch_processing_result):
        for p in self.policies:
            p.update_state(batch_processing_result)

    def need_shutdown(self):
        for p in self.policies:
            if not p.need_shutdown():
                return False
        return True

    def __repr__(self):
        repr_policies = ", ".join([repr(p) for p in self.policies])
        return "AndShutdown({})".format(repr_policies)


NEVER_SHUTDOWN = NeverShutdown()
ASAP_SHUTDOWN = IdleShutdown(0)



from contextlib import contextmanager
from threading import local


class SQSContext(local):
    """
    Context which is implicitly passed to the worker via the job message. Can
    be used there for logging or profiling purposes, for example.

    Usage example.


        @sqs.processor('q1', 'hello_world', pass_context=True)
        def hello_world(username=None, context=None):
            print(f'Hello {username} from {context["remote_addr"]}')

        with sqs.context(remote_addr='127.0.0.1'):
            hello_world.delay(username='Foo')

        sqs.process_batch('q1')

    Alternatively, you can set the context like this.

        sqs.context['remote_addr'] = '127.0.0.1'
        hello_world.delay(username='Foo')

    And then, when the context needs to be cleared:

        sqs.context.clear()

    In a web application you usually call it at the end of the processing
    of the web request.
    """

    @classmethod
    def from_dict(cls, kwargs):
        """
        Create context from the dict
        """
        return cls(**kwargs)

    def __init__(self, **kwargs):
        """
        Create a new context and populate it with keys and values from kwargs
        """
        self._context = {}
        self.set(**kwargs)

    def set(self, **kwargs):
        """
        Clean up current context and replace it with values from kwargs
        """
        self._context = kwargs.copy()

    def update(self, **kwargs):
        """
        Extend current context with values from kwargs
        """
        self._context.update(kwargs)

    def clear(self):
        """
        Remove all values from the context
        """
        self._context.clear()

    def to_dict(self):
        """
        Convert context to dictionary
        """
        return self._context.copy()

    def get(self, key, default=None):
        """
        Get value by key from the context
        """
        return self._context.get(key, default)

    def __getitem__(self, item):
        """
        Dict API emulation. Get value by key from the context
        """
        return self._context[item]

    def __setitem__(self, key, value):
        """
        Dict API emulation. Set value by key
        """
        self._context[key] = value

    @contextmanager
    def __call__(self, **kwargs):
        """
        Run the context manager. Temporarily extend current context with
        kwargs until the end of the context manager. On exit from the
        with-statement restore back the original context.
        """
        orig_context = self._context.copy()
        try:
            self.update(**kwargs)
            yield self
        finally:
            self.set(**orig_context)

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self._context)



import logging
from functools import partial
from typing import TYPE_CHECKING, Callable, Optional

import attr

from sqs_workers import codecs
from sqs_workers.context import SQSContext
from sqs_workers.utils import adv_validate_arguments

if TYPE_CHECKING:
    from sqs_workers.queue import GenericQueue

logger = logging.getLogger(__name__)

DEFAULT_CONTEXT_VAR = "context"


@attr.s
class Processor(object):
    """
    Processor which calls its function for each incoming message.

    fn() accepts a single parameter "message" which is a decoded body of an
    SQS message
    """

    queue = attr.ib()  # type: GenericQueue
    fn = attr.ib(default=None)  # type: Optional[Callable]
    job_name = attr.ib(default="")  # type: str
    pass_context = attr.ib(default=False)  # type: bool
    context_var = attr.ib(default=DEFAULT_CONTEXT_VAR)  # type: str

    @classmethod
    def maker(cls, **kwargs):
        return partial(cls, **kwargs)

    def process_message(self, message):
        extra = {
            "message_id": message.message_id,
            "queue_name": self.queue.name,
            "job_name": self.job_name,
        }
        logger.debug("Process {queue_name}.{job_name}".format(**extra), extra=extra)

        try:
            content_type = get_job_content_type(message)
            extra["job_content_type"] = content_type
            codec = codecs.get_codec(content_type)
            job_kwargs = codec.deserialize(message.body)
            job_context = get_job_context(message, codec)
            self.process(job_kwargs, job_context)
        except Exception:
            logger.exception(
                "Error while processing {queue_name}.{job_name}".format(**extra),
                extra=extra,
            )
            return False
        else:
            return True

    def process(self, job_kwargs, job_context):
        effective_kwargs = job_kwargs.copy()
        if self.pass_context:
            effective_kwargs[self.context_var] = job_context
        return call_handler(self.fn, effective_kwargs)

    def copy(self, **kwargs):
        """
        Create a new instance of the processor, optionally updating
        arguments of the constructor from update_kwargs
        """
        return attr.evolve(self, **kwargs)


def get_job_content_type(job_message):
    attrs = job_message.message_attributes or {}
    return (attrs.get("ContentType") or {}).get("StringValue")


def get_job_context(job_message, codec):
    attrs = job_message.message_attributes or {}
    serialized = (attrs.get("JobContext") or {}).get("StringValue")
    if not serialized:
        return SQSContext()
    deserialized = codec.deserialize(serialized)
    return SQSContext.from_dict(deserialized)


def call_handler(fn, kwargs):
    try:
        handler_args, handler_kwargs = adv_validate_arguments(fn, [], kwargs)
    except TypeError:
        # it may happen, if "fn" is not a function (but
        # a mock object, for example)
        handler_args, handler_kwargs = [], kwargs
    fn(*handler_args, **handler_kwargs)



import logging
import warnings
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional

import attr

from sqs_workers import DEFAULT_BACKOFF, codecs
from sqs_workers.async_task import AsyncTask
from sqs_workers.backoff_policies import BackoffPolicy
from sqs_workers.core import BatchProcessingResult, get_job_name
from sqs_workers.processors import DEFAULT_CONTEXT_VAR, Processor
from sqs_workers.shutdown_policies import NEVER_SHUTDOWN

DEFAULT_MESSAGE_GROUP_ID = "default"

if TYPE_CHECKING:
    from sqs_workers import SQSEnv

logger = logging.getLogger(__name__)


@attr.s
class GenericQueue(object):
    env = attr.ib(repr=False)  # type: SQSEnv
    name = attr.ib()  # type: str
    backoff_policy = attr.ib(default=DEFAULT_BACKOFF)  # type: BackoffPolicy
    _queue = attr.ib(repr=False, default=None)

    @classmethod
    def maker(cls, **kwargs):
        return partial(cls, **kwargs)

    def process_queue(self, shutdown_policy=NEVER_SHUTDOWN, wait_second=10):
        """
        Run worker to process one queue in the infinite loop
        """
        logger.debug(
            "Start processing queue {}".format(self.name),
            extra={
                "queue_name": self.name,
                "wait_seconds": wait_second,
                "shutdown_policy": repr(shutdown_policy),
            },
        )
        while True:
            result = self.process_batch(wait_seconds=wait_second)
            shutdown_policy.update_state(result)
            if shutdown_policy.need_shutdown():
                logger.debug(
                    "Stop processing queue {}".format(self.name),
                    extra={
                        "queue_name": self.name,
                        "wait_seconds": wait_second,
                        "shutdown_policy": repr(shutdown_policy),
                    },
                )
                break

    def process_batch(self, wait_seconds=0):
        # type: (int) -> BatchProcessingResult
        """
        Process a batch of messages from the queue (10 messages at most), return
        the number of successfully processed messages, and exit
        """
        queue = self.get_queue()
        messages = self.get_raw_messages(wait_seconds)
        result = BatchProcessingResult(self.name)

        for message in messages:
            success = self.process_message(message)
            result.update_with_message(message, success)
            if success:
                entry = {
                    "Id": message.message_id,
                    "ReceiptHandle": message.receipt_handle,
                }
                queue.delete_messages(Entries=[entry])
            else:
                timeout = self.backoff_policy.get_visibility_timeout(message)
                message.change_visibility(VisibilityTimeout=timeout)
        return result

    def process_message(self, message):
        # type: (Any) -> bool
        """
        Process single message.

        Return True if processing went successful
        """
        raise NotImplementedError()

    def get_raw_messages(self, wait_seconds):
        """
        Return raw messages from the queue, addressed by its name
        """
        kwargs = {
            "WaitTimeSeconds": wait_seconds,
            "MaxNumberOfMessages": 10,
            "MessageAttributeNames": ["All"],
            "AttributeNames": ["All"],
        }
        queue = self.get_queue()
        return queue.receive_messages(**kwargs)

    def drain_queue(self, wait_seconds=0):
        """
        Delete all messages from the queue without calling purge().
        """
        queue = self.get_queue()
        deleted_count = 0
        while True:
            messages = self.get_raw_messages(wait_seconds)
            if not messages:
                break
            entries = [
                {"Id": msg.message_id, "ReceiptHandle": msg.receipt_handle}
                for msg in messages
            ]
            queue.delete_messages(Entries=entries)
            deleted_count += len(messages)
        return deleted_count

    def get_sqs_queue_name(self):
        """
        Take "high-level" (user-visible) queue name and return SQS
        ("low level") name by simply prefixing it. Used to create namespaces
        for different environments (development, staging, production, etc)
        """
        return self.env.get_sqs_queue_name(self.name)

    def get_queue(self):
        """
        Helper function to return queue object.
        """
        if self._queue is None:
            self._queue = self.env.sqs_resource.get_queue_by_name(
                QueueName=self.get_sqs_queue_name()
            )
        return self._queue


@attr.s
class RawQueue(GenericQueue):
    processor = attr.ib(default=None)  # type: Optional[Callable]

    def raw_processor(self):
        """
        Decorator to assign processor to handle jobs from the raw queue

        Usage example:

        cron = sqs.queue('cron')

        @cron.raw_processor()
        def process(message):
            print(message.body)
        """

        def func(processor):
            return self.connect_raw_processor(processor)

        return func

    def connect_raw_processor(self, processor):
        """
        Assign raw processor (a function) to handle jobs.
        """
        extra = {
            "queue_name": self.name,
            "processor_name": processor.__module__ + "." + processor.__name__,
        }
        logger.debug(
            "Connect {queue_name} to " "processor {processor_name}".format(**extra),
            extra=extra,
        )
        self.processor = processor

    def add_raw_job(
        self,
        message_body,  # type: str
        delay_seconds=0,  # type: int
        deduplication_id=None,  # type: str
        group_id=DEFAULT_MESSAGE_GROUP_ID,  # type: str
    ):
        """
        Add raw message to the queue
        """

        kwargs = {
            "MessageBody": message_body,
            "DelaySeconds": delay_seconds,
            "MessageAttributes": {},
        }
        if self.name.endswith(".fifo"):
            # if queue name ends with .fifo, then according to the AWS specs,
            # it's a FIFO queue, and requires group_id.
            # Otherwise group_id can be set to None
            if group_id is None:
                group_id = DEFAULT_MESSAGE_GROUP_ID
            if deduplication_id is not None:
                kwargs["MessageDeduplicationId"] = str(deduplication_id)
            if group_id is not None:
                kwargs["MessageGroupId"] = str(group_id)

        ret = self.get_queue().send_message(**kwargs)
        return ret["MessageId"]

    def process_message(self, message):
        # type: (Any) -> bool
        """
        Process single message.

        Return True if processing went successful
        """
        extra = {"message_id": message.message_id, "queue_name": self.name}
        if not self.processor:
            logger.warning(
                "No processor set for {queue_name}".format(**extra), extra=extra
            )

        logger.debug("Process {queue_name}.{message_id}".format(**extra), extra=extra)

        try:
            self.processor(message)
        except Exception:
            logger.exception(
                "Error while processing {queue_name}.{message_id}".format(**extra),
                extra=extra,
            )
            return False
        else:
            return True


@attr.s
class JobQueue(GenericQueue):

    processors = attr.ib(factory=dict)  # type: Dict[str, Processor]

    def processor(self, job_name, pass_context=False, context_var=DEFAULT_CONTEXT_VAR):
        """
        Decorator to assign processor to handle jobs with the name job_name
        from the queue queue_name

        Usage example:

        queue = sqs.queue('q1')

        @queue.processor('say_hello')
        def say_hello(name):
            print("Hello, " + name)

        Then you can add messages to the queue by calling ".delay" attribute
        of a newly created object:

        >>> say_hello.delay(name='John')

        Here, instead of calling function locally, it will be pushed to the
        queue (essentially, mimics the non-blocking call of the function).

        If you still want to call function locally, you can call

        >>> say_hello(name='John')
        """

        def fn(processor):
            return self.connect_processor(
                job_name, processor, pass_context, context_var
            )

        return fn

    def connect_processor(
        self, job_name, processor, pass_context=False, context_var=DEFAULT_CONTEXT_VAR
    ):
        """
        Assign processor (a function) to handle jobs with the name job_name
        from the queue queue_name
        """
        extra = {
            "queue_name": self.name,
            "job_name": job_name,
            "processor_name": processor.__module__ + "." + processor.__name__,
        }
        logger.debug(
            "Connect {queue_name}.{job_name} to "
            "processor {processor_name}".format(**extra),
            extra=extra,
        )
        self.processors[job_name] = self.env.processor_maker(
            queue=self,
            fn=processor,
            pass_context=pass_context,
            context_var=context_var,
            job_name=job_name,
        )
        return AsyncTask(self, job_name, processor)

    def get_processor(self, job_name):
        """
        Helper function to return a processor for the queue
        """
        return self.processors.get(job_name)

    def add_job(
        self,
        job_name,
        _content_type=codecs.DEFAULT_CONTENT_TYPE,
        _delay_seconds=None,
        _deduplication_id=None,
        _group_id=None,
        **job_kwargs
    ):
        """
        Add job to the queue. The body of the job will be converted to the text
        with one of the codes (by default it's "pickle")
        """
        codec = codecs.get_codec(_content_type)
        message_body = codec.serialize(job_kwargs)
        job_context = codec.serialize(self.env.context.to_dict())
        return self.add_raw_job(
            job_name,
            message_body,
            job_context,
            _content_type,
            _delay_seconds,
            _deduplication_id,
            _group_id,
        )

    def add_raw_job(
        self,
        job_name,
        message_body,
        job_context,
        content_type,
        delay_seconds,
        deduplication_id,
        group_id,
    ):
        """
        Low-level function to put message to the queue
        """
        # if queue name ends with .fifo, then according to the AWS specs,
        # it's a FIFO queue, and requires group_id.
        # Otherwise group_id can be set to None
        if group_id is None and self.name.endswith(".fifo"):
            group_id = DEFAULT_MESSAGE_GROUP_ID

        queue = self.get_queue()
        kwargs = {
            "MessageBody": message_body,
            "MessageAttributes": {
                "ContentType": {"StringValue": content_type, "DataType": "String"},
                "JobContext": {"StringValue": job_context, "DataType": "String"},
                "JobName": {"StringValue": job_name, "DataType": "String"},
            },
        }
        if delay_seconds is not None:
            kwargs["DelaySeconds"] = int(delay_seconds)
        if deduplication_id is not None:
            kwargs["MessageDeduplicationId"] = str(deduplication_id)
        if group_id is not None:
            kwargs["MessageGroupId"] = str(group_id)
        ret = queue.send_message(**kwargs)
        return ret["MessageId"]

    def process_message(self, message):
        # type: (Any) -> bool
        """
        Process single message.

        Return True if processing went successful
        """
        job_name = get_job_name(message)
        processor = self.get_processor(job_name)
        if processor:
            return processor.process_message(message)
        else:
            return self.process_message_fallback(message)

    def process_message_fallback(self, message):
        warnings.warn(
            "Error while processing {}.{}".format(self.name, get_job_name(message))
        )
        return False

    def copy_processors(self, dst_queue):
        # type: (JobQueue) -> None
        """
        Copy processors from self to dst_queue. Can be helpful to process d
        ead-letter queue with processors from the main queue.

        Usage example.

        sqs = SQSEnv()
        ...
        foo = sqs.queue('foo')
        foo_dead = sqs.queue('foo_dead')
        foo.copy_processors('foo_dead')
        foo_dead.process_queue(shutdown_policy=IdleShutdown(10))

        Here the queue "foo_dead" will be processed with processors from the
        queue "foo".
        """
        for job_name, processor in self.processors.items():
            dst_queue.processors[job_name] = processor.copy(queue=dst_queue)



import json
import logging
from typing import Any

logger = logging.getLogger(__name__)


class BatchProcessingResult(object):
    succeeded = None
    failed = None

    def __init__(self, queue_name, succeeded=None, failed=None):
        self.queue_name = queue_name
        self.succeeded = succeeded or []
        self.failed = failed or []

    def update_with_message(self, message, success):
        # type: (Any, bool) -> None
        """
        Update processing result with a message.
        """
        if success:
            self.succeeded.append(message)
        else:
            self.failed.append(message)

    def succeeded_count(self):
        return len(self.succeeded)

    def failed_count(self):
        return len(self.failed)

    def total_count(self):
        return self.succeeded_count() + self.failed_count()

    def __repr__(self):
        return "<BatchProcessingResult/%s/%s/%s>" % (
            self.queue_name,
            self.succeeded_count(),
            self.failed_count(),
        )


class RedrivePolicy(object):
    """
    Redrive Policy for SQS queues

    See for more details:
    https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-dead-letter-queues.html
    """

    def __init__(self, sqs_env, dead_letter_queue_name, max_receive_count):
        self.sqs_env = sqs_env
        self.dead_letter_queue_name = dead_letter_queue_name
        self.max_receive_count = max_receive_count

    def __json__(self):
        queue = self.sqs_env.sqs_resource.get_queue_by_name(
            QueueName=self.sqs_env.get_sqs_queue_name(self.dead_letter_queue_name)
        )
        target_arn = queue.attributes["QueueArn"]
        # Yes, it's double-encoded JSON :-/
        return json.dumps(
            {
                "deadLetterTargetArn": target_arn,
                "maxReceiveCount": str(self.max_receive_count),
            }
        )


def get_job_name(message):
    attrs = message.message_attributes or {}
    return (attrs.get("JobName") or {}).get("StringValue")
